#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""CB-210 CLIコマンド

usage:
$ python3 cb210cli.py (Linux)
> python cb210cli.py (Windows)

"""
import sys
import argparse
import time
import tempfile
from datetime import datetime

import serial_tranceiver

class CB210Cli():
    """CB-210 Serial I/F"""

    TIMEOUT_DEFAULT = 5
    RETRY_DEFAULT = 3

    stx = b'\x02'
    etx = b'\x03'
    eot = b'\x04'
    ack = b'\x06'
    nak = b'\x15'

    def __init__(self, port, timeout, retry, verbose):
        if port is None:
            self.serial = None
        else:
            self._retry = retry
            self.serial = serial_tranceiver.SerialTranceiver(
                port, eol=self.eot, timeout_rw=timeout)
            self._verbose = verbose

    def dump(self, filename):
        """
        CB-210の設定をファイルに記録する。
        """

        try:
            with open(filename, mode='x') as f:
                # self._send()の中で3回リトライするので、_send()が例外発生したらリカバリは
                # 不可能と思われるので例外を素通しして良い。
                self._send("HELLO")

                self._write_header(f)
                # TODO: CB-210に設定ダンプコマンドを用意する。
                # v1.0.0時点では1つづつ問い合わせコマンドを実行する。
                self._dump_1_0_0(f)

            print('success')

        except FileExistsError:
            raise Exception('File exists: \'{0}\''.format(filename))

    def restore(self, filename, save_after_restore):
        """
        ファイルにある設定をCB-210に設定する。
        """

        try:
            with open(filename) as f:
                self._send("HELLO")
                for line in f:
                    line = line.rstrip(' \r\n')
                    comment_index = line.find('#')
                    if comment_index >= 0:
                        line = line[:comment_index]
                    if len(line) == 0:
                        continue

                    self._send(line)

            if save_after_restore:
                self._send("CCFGSAVE")

            print('success')

        except FileNotFoundError:
            raise Exception('No such file or directory: \'{0}\''.format(filename))


    def interactive(self):
        """
        対話モード
        """
        while True:
            cmd = input('Input command, q:quit) > ')
            if cmd == 'q':
                break

            try:
                self._send_command(cmd)
                while True:
                    resp = self._receive_response()
                    if isinstance(resp, Notify):
                        # 通知を受信。表示してまたレスポンスの受信待ち。
                        msg = 'Received notify {0}, param = \"{1}\"'\
                                .format(resp.command, resp.param)
                        print(msg)
                        continue

                    # レスポンスが来た。
                    # レスポンス内のコマンド種別が合っているか判定。
                    len_core = min(len(cmd), len(resp.command))

                    if cmd[:len_core] != resp.command[:len_core]:
                        # 違うコマンドのレスポンスが来た。表示してまたレスポンスの受信待ち。
                        msg = 'Received different command response: command = {}, {}, param = \"{}\"'\
                                .format(resp.command[:len_core], ('ACK' if resp.acknak else 'NAK'),
                                        resp.param)
                        print(msg)
                        continue

                    # 送ったコマンドのレスポンスが来た。受信終了。
                    msg = 'Received {0}, param = \"{1}\"'\
                            .format(('ACK' if resp.acknak else 'NAK'),
                                    resp.param)
                    print(msg)
                    break

            except Exception as err:
                print("Exception: {0}".format(err))

    def _write_header(self, fp):
        """
        ファイルに説明用のヘッダを書き込む
        """
        ctime = datetime.now().ctime()
        fp.write("# CB-210 config: {0}\n".format(ctime))

    def _dump_1_0_0(self, fp):
        """
        設定をファイルにダンプする。全ての設定問い合わせコマンドを送る。
        """

        # 問い合わせコマンド一覧。(v1.0.0時点)
        # v1.1.0以降は、CB-210側に設定ダンプコマンドを用意する予定なので、今後は使わない予定。
        COMMANDS_TO_DUMP = ['SSELC1', 'SORSTT1', 'SIT', 'SWARU1', 'SWASH1',
            'SWASK1', 'SDI1', 'SDI2', 'SDO1']
        COMMANDS_TO_DUMP_WITH_PARAM = {
            'SOTIM1': range(1, 6)
        }
        # パラメータ無しの問い合わせ
        for cmd in COMMANDS_TO_DUMP:
            resp = self._send(cmd)
            if resp:
                cmd_setting = self._convert_query_to_set(cmd)
                fp.write('{0}{1}\n'.format(cmd_setting, resp))

        # パラメータ有りの問い合わせ
        for cmd, params in COMMANDS_TO_DUMP_WITH_PARAM.items():
            for param in params:
                param_str = str(param)
                resp = self._send(cmd + param_str)
                if resp:
                    cmd_setting = self._convert_query_to_set(cmd)
                    fp.write('{0}{1}{2}\n'.format(cmd_setting, param, resp))

    @staticmethod
    def _convert_query_to_set(cmd):
        return cmd.replace('S', 'C', 1)

    def _send(self, command):
        """
        CB-210にコマンドを送る。retry回数繰り返してもACKが返って来なかったら例外発生。
        非対話モードで使うことを想定している。
        Returns: param部分(str)
        Exception: NAKが帰ってきたらExceptionが発生する。
        """
        for i in range(self._retry):
            try:
                self._send_command(command)

                while True:
                    resp = self._receive_response()

                    if isinstance(resp, Notify):
                        print("Notice: notify received: {0}".format(resp.command))
                        continue

                    len_core = min(len(command), len(resp.command))

                    if command[:len_core] == resp.command[:len_core]:
                        break
                    else:
                        print("Warning: Invalid response. cmd={0}".format(resp.command))

                if not resp.acknak:
                    raise Exception("Warning: Received NAK")

                return resp.param
            except CheckCodeException as e:
                # チェックコード不一致はデータ化けなのでもう一度実行してみる
                print("Warning: {0}".format(e))
            except TimeoutError as e:
                # タイムアウトはもう一度実行してみる
                print("Warning: {0}".format(e))
            except Exception as e:
                # それ以外の例外は致命的なので処理を中断する
                raise e
        else:
            raise Exception("Retry exceeded.")

    def _calculate_checkcode(self, command):
        """チェックコードを求める

        データ部を1バイト単位で加算し、その結果の下位2バイトの2の補数の
        2文字のHEXコードをASCIIコード化したデータがチェックコードになる。

        Args:
            command(bytes): コマンド文字列

        Returns:
            チェックコード(bytes)

        """
        total = 0
        for char in command:
            total += char
        lower2hex = total & 0xFF
        complement_2 = ~lower2hex + 0x01
        # 2の補数を16進表現した文字の下位2桁を1文字ずつキャラクタに変換し、
        # そのキャラクタのASCIIコードを返す
        upper = (complement_2 >> 4) & 0x0F
        lower = complement_2 & 0x0F
        str_upper = '{:X}'.format(upper)
        str_lower = '{:X}'.format(lower)
        ckbyte1 = str_upper.encode()
        ckbyte2 = str_lower.encode()
        return ckbyte1 + ckbyte2

    def _send_command(self, command):
        """コマンドを送信する

        Args:
            command(str): コマンド文字列

        Returns:
            None

        """

        if isinstance(command, str):
            command = command.encode()

        ckcode = self._calculate_checkcode(command)

        send_bytes = self.stx + command + self.etx + ckcode + self.eot

        if self.serial is None:
            print("command is {0}".format(send_bytes))
            return

        if self._verbose:
            print('Send: {0}'.format(send_bytes), file=sys.stderr)

        self.serial.write_string(send_bytes)

    def _receive_response(self):
        """送信したコマンドの応答や通知を受け取る

        Args:
            なし

        Returns:
            CommandResponseまたはNofityのインスタンス

        """
        if self.serial is None:
            return b''

        data = self.serial.readline()

        if self._verbose:
            print('Recv: {0}'.format(data), file=sys.stderr)

        stx = data.find(self.stx)
        etx = data.find(self.etx)
        if stx < 0:
            raise Exception("stx not found")

        if etx < 0:
            raise Exception("etx not found")

        respdata = data[stx+len(self.stx):etx]
        checkcode_recv = data[etx+1:-1]

        ckcode = self._calculate_checkcode(respdata)

        if checkcode_recv != ckcode:
            raise CheckCodeException(checkcode_recv, ckcode)

        if respdata[0:1] == b'A':
            # 通知を受信
            return Notify(respdata.decode("ascii"))

        else:
            # コマンドの返答を受信
            # respdataをパースする
            ack = respdata.find(self.ack)
            nak = respdata.find(self.nak)
            if ack >= 0:
                acknak_index = ack
                acknak = True
            elif nak >= 0:
                acknak_index = nak
                acknak = False
            else:
                raise Exception("ACK or NAK not found. respdata={0}".format(respdata))

            command = respdata[:acknak_index].decode("ascii")
            param = respdata[acknak_index+1:].decode("ascii")
            return CommandResponse(command, acknak, param)

class Response:
    def __init__(self, command, param=None):
        self._command = command
        self._param = param

    @property
    def command(self):
        return self._command

    @property
    def param(self):
        return self._param

class CommandResponse(Response):
    def __init__(self, command, acknak, param):
        super().__init__(command, param)
        self._acknak = acknak

    @property
    def acknak(self):
        return self._acknak

class Notify(Response):
    def __init__(self, command):
        super().__init__(command, None)

class CheckCodeException(Exception):
    """
    チェックコードが不一致の時に発生する例外。
    """
    def __init__(self, recv, calc):
        msg = "Check code mismatch. recv = {0}, calc = {1}"\
                .format(recv, calc)
        super().__init__(msg)

def parse_args():
    parser = argparse.ArgumentParser(description="CB-210 config tool")
    parser.add_argument('device', help='Serial Device'
        '(Linux: /dev/ttyUSB0 etc. Win: com1 etc.)')
    parser.add_argument('-t', '--timeout', metavar='SECS',
        default=CB210Cli.TIMEOUT_DEFAULT, type=int,
        help='Timeout(sec) (Default: {0})'.format(CB210Cli.TIMEOUT_DEFAULT))
    parser.add_argument('--retry', metavar='TIMES',
        default=CB210Cli.RETRY_DEFAULT, type=int,
        help='Retry count (Default: {0})'.format(CB210Cli.RETRY_DEFAULT))
    parser.add_argument('--save', action='store_true',
        help='Send "save config" command to CB-210 after restore (Default: no)')
    parser.add_argument('-v', '--verbose', action='store_true',
        help='Verbose mode (Default: no)')

    # Mode (dump or restore or normal)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-d', '--dump', metavar='FILENAME',
        help='Dump config to file')
    group.add_argument('-r', '--restore', metavar='FILENAME',
        help='Restore config from file')

    return parser.parse_args()

def main():
    """主処理"""

    try:
        args = parse_args()
        cli = CB210Cli(args.device, args.timeout, args.retry, args.verbose)
    except Exception as err:
        print(err)
        sys.exit(1)

    print("CB-210 config tool")

    try:
        if args.dump:
            print("Dump config to", args.dump)
            cli.dump(args.dump)
        elif args.restore:
            print("Restore config from", args.restore)
            cli.restore(args.restore, args.save)
        else:
            cli.interactive()

        sys.exit(0)
    except Exception as e:
        print(e)
        sys.exit(1)

if __name__ == '__main__':
    main()
